//
//	GSMacOSXFramework.cpp
//	 2001 Kyle Hammond
//	kyle@codeblazer.com
//	Use at your own risk.  All rights reserved.  Do not distribute without permission.
//
#include "GSMacOSXFramework.h"

static OSStatus local_loadFrameworkBundle( CFStringRef framework, CFBundleRef *bundlePtr );

GSMacOSXFramework::GSMacOSXFramework( const char *inFrameworkName ) : mFrameworkBundle( nil )
{	OSErr			err;
	CFMutableStringRef	tempString = nil;

	tempString = CFStringCreateMutable( kCFAllocatorSystemDefault, 30 );
	if ( tempString )
		CFStringAppendCString( tempString, inFrameworkName, kCFStringEncodingMacRoman );
	if ( tempString )
		CFStringAppendCString( tempString, ".framework", kCFStringEncodingMacRoman );
	err = local_loadFrameworkBundle( tempString, &mFrameworkBundle );
	if ( err )
		mFrameworkBundle = nil;

	if ( tempString )
		CFRelease( tempString );
}

GSMacOSXFramework::~GSMacOSXFramework( )
{	if ( mFrameworkBundle ) {
		CFBundleUnloadExecutable( mFrameworkBundle );

		CFRelease( mFrameworkBundle );
	}
}

#pragma mark -

static OSStatus local_loadFrameworkBundle( CFStringRef framework, CFBundleRef *bundlePtr )
{	OSStatus 	err;
	FSRef 	frameworksFolderRef;
	CFURLRef	baseURL;
	CFURLRef	bundleURL;

	if ( bundlePtr == nil || framework == nil )
		return paramErr;

	*bundlePtr = nil;
	
	baseURL = nil;
	bundleURL = nil;
	
	err = FSFindFolder( kOnAppropriateDisk, kFrameworksFolderType, true, &frameworksFolderRef );
	if (err == noErr) {
		baseURL = CFURLCreateFromFSRef( kCFAllocatorSystemDefault, &frameworksFolderRef );
		if (baseURL == nil) {
			err = coreFoundationUnknownErr;
		}
	}
	if (err == noErr) {
		bundleURL = CFURLCreateCopyAppendingPathComponent(kCFAllocatorSystemDefault, baseURL, framework, false);
		if (bundleURL == nil) {
			err = coreFoundationUnknownErr;
		}
	}
	if (err == noErr) {
		*bundlePtr = CFBundleCreate(kCFAllocatorSystemDefault, bundleURL);
		if (*bundlePtr == nil) {
			err = coreFoundationUnknownErr;
		}
	}
	if (err == noErr) {
	    if ( !CFBundleLoadExecutable( *bundlePtr ) ) {
			err = coreFoundationUnknownErr;
	    }
	}

	// Clean up.
	
	if ( err != noErr && *bundlePtr != nil ) {
		CFRelease( *bundlePtr );
		*bundlePtr = nil;
	}

	if ( bundleURL != nil )
		CFRelease( bundleURL );

	if ( baseURL != nil )
		CFRelease( baseURL );

	return err;
}
