//
//	GSSerialIOKitFramework.h
//	 2001 Kyle Hammond
//	kyle@codeblazer.com
//	Use at your own risk.  All rights reserved.  Do not distribute without permission.
//
#include "GSMacOSXFramework.h"

// From <mach/ppc/kern_return.h>
typedef	int		kern_return_t;
#define KERN_SUCCESS			0
#define KERN_FAILURE			5

// From <mach/ppc/vm_types.h>
typedef unsigned int	natural_t;

// From <mach/port.h>
typedef natural_t		port_name_t;
typedef port_name_t		port_t;
typedef port_t			mach_port_t;
#define MACH_PORT_NULL		0  /* intentional loose typing */

// From <IOKit/IOTypes.h>
typedef mach_port_t		io_object_t;
typedef io_object_t		io_iterator_t;
typedef UInt32			IOOptionBits;
typedef io_object_t		io_registry_entry_t;

// From <IOKit/serial/IOSerialKey.h>
#define kIOSerialBSDServiceValue	"IOSerialBSDClient"
#define kIOSerialBSDTypeKey		"IOSerialBSDClientType"
#define kIOSerialBSDRS232Type	"IORS232SerialStream"
#define kIOTTYDeviceKey			"IOTTYDevice"
#define kIOCalloutDeviceKey		"IOCalloutDevice"

// From <syslimits.h>
#define	PATH_MAX		 1024	/* max bytes in pathname */

// From <param.h>
#define	MAXPATHLEN	PATH_MAX

// From <termios.h>
#define	NCCS		20
typedef unsigned long	tcflag_t;
typedef unsigned char	cc_t;
typedef long			speed_t;	/* XXX should be unsigned long */

struct termios {
	tcflag_t	c_iflag;	/* input flags */
	tcflag_t	c_oflag;	/* output flags */
	tcflag_t	c_cflag;	/* control flags */
	tcflag_t	c_lflag;	/* local flags */
	cc_t		c_cc[ NCCS ];	/* control chars */
	speed_t	c_ispeed;	/* input speed */
	speed_t	c_ospeed;	/* output speed */
};


class GSSerialIOKitFramework : public GSMacOSXFramework {
public:
	GSSerialIOKitFramework( );

	// From <IOKitLib.h>
	kern_return_t	IOMasterPort( mach_port_t bootstrapPort, mach_port_t *masterPort );

	io_object_t	IOIteratorNext( io_iterator_t inIterator );

	CFTypeRef	IORegistryEntryCreateCFProperty( io_registry_entry_t entry, CFStringRef key,
							CFAllocatorRef allocator, IOOptionBits options );

	CFMutableDictionaryRef IOServiceMatching( const char *inName );

	kern_return_t	IOServiceGetMatchingServices( mach_port_t masterPort, CFDictionaryRef matching,
				io_iterator_t *existing );

	kern_return_t	IOObjectRelease( io_object_t object );

	virtual ~GSSerialIOKitFramework( ) { }

private:
	typedef kern_return_t (*IOMasterPortPtr)( mach_port_t bootstrapPort, mach_port_t *masterPort );
	typedef io_object_t (*IOIteratorNextPtr)( io_iterator_t inIterator );
	typedef CFTypeRef (*IORegistryEntryCreateCFPropertyPtr)( io_registry_entry_t entry, CFStringRef key,
							CFAllocatorRef allocator, IOOptionBits options );
	typedef CFMutableDictionaryRef (*IOServiceMatchingPtr)( const char *inName );
	typedef kern_return_t (*IOServiceGetMatchingServicesPtr)( mach_port_t masterPort, CFDictionaryRef matching,
				io_iterator_t *existing );
	typedef kern_return_t (*IOObjectReleasePtr)( io_object_t object );

	IOMasterPortPtr					mIOMasterPort;
	IOIteratorNextPtr					mIOIteratorNext;
	IORegistryEntryCreateCFPropertyPtr	mIORegistryEntryCreateCFProperty;
	IOServiceMatchingPtr				mIOServiceMatching;
	IOServiceGetMatchingServicesPtr		mIOServiceGetMatchingServices;
	IOObjectReleasePtr					mIOObjectRelease;

	GSSerialIOKitFramework( const GSSerialIOKitFramework & );
};