//
//	GSSystemFramework.h
//	 2001 Kyle Hammond
//	kyle@codeblazer.com
//	Use at your own risk.  All rights reserved.  Do not distribute without permission.
//
#include "GSMacOSXFramework.h"

// From <Kernal/fcntl.h>
#define	F_SETFL		4		/* set file status flags */

#define	O_RDWR		0x0002		/* open for reading and writing */

#define	O_NOCTTY	0		/* don't assign controlling terminal */

#define	O_NONBLOCK	0x0004		/* no delay */

#define	O_NDELAY		O_NONBLOCK	/* compat */

// From <Kernal/termios.h>
#define	ISIG		0x00000080	/* enable signals INTR, QUIT, [D]SUSP */
#define	ICANON		0x00000100	/* canonicalize input lines */

#define VMIN		16	/* !ICANON */
#define VTIME		17	/* !ICANON */

#define	TCSANOW		0		/* make change immediate */

#define	OPOST		0x00000001	/* enable following output processing */

#define     CS5		    0x00000000	    /* 5 bits (pseudo) */
#define     CS6		    0x00000100	    /* 6 bits */
#define     CS7		    0x00000200	    /* 7 bits */
#define     CS8		    0x00000300	    /* 8 bits */
#define CSTOPB		0x00000400	/* send 2 stop bits */
#define CREAD		0x00000800	/* enable receiver */
#define PARENB		0x00001000	/* parity enable */
#define PARODD		0x00002000	/* odd parity, else even */
#define CLOCAL		0x00008000	/* ignore modem status lines */

#define	ECHOE		0x00000002	/* visually erase chars */
#define ECHO		0x00000008	/* enable echoing */

class GSSystemFramework : public GSMacOSXFramework {
public:
	GSSystemFramework( );

	// From <termios.h>
	int	tcgetattr (int, struct termios *);
	int	tcsetattr (int, int, const struct termios *);

	// From <fcntl.h>
	int	open (const char *, int, ...);
	int	fcntl (int, int, ...);
	int	close ( int );
	int	write ( int, void *, __std(size_t) );
	int	read ( int, void *, __std(size_t) );

	// From <errno.h>
	int	errno( void ) const;

	virtual ~GSSystemFramework( ) { }

private:
	typedef int (*tcgetattrPtr) (int, struct termios *);
	typedef int (*tcsetattrPtr) (int, int, const struct termios *);
	typedef int (*openPtr) (const char *, int, ...);
	typedef int (*fcntlPtr) (int, int, ...);
	typedef int (*closePtr) ( int );
	typedef int (*writePtr) ( int, void *, __std(size_t) );
	typedef int (*readPtr) ( int, void *, __std(size_t) );
	typedef int *(*__errorPtr) ( void );

	tcgetattrPtr	mtcgetattr;
	tcsetattrPtr	mtcsetattr;
	openPtr		mopen;
	fcntlPtr		mfcntl;
	closePtr		mclose;
	writePtr		mwrite;
	readPtr		mread;
	__errorPtr	merror;

	GSSystemFramework( const GSSystemFramework & );
};